#!/usr/bin/perl -w

# $Id: LZW.pm,v 1.15 2003/02/13 04:32:42 jim Exp $
# Based on u6decode.c and publically-available LZW
# decompression docs.

package U6::LZW;
BEGIN {
	use Exporter;
	our ($VERSION, @ISA, @EXPORT, @EXPORT_OK);
	$VERSION = do { my @r = (q$Revision: 1.15 $ =~ /\d+/g); sprintf "%d."."%02d" x $#r, @r };
	@ISA = qw(Exporter);
	@EXPORT    = qw();
	@EXPORT_OK = qw(lzw_decompress_file);
}

use strict;
use Fcntl qw(:seek);
my $true = 1;
my $false = 1;
my $DEFAULT_CW_SIZE = 9;
my $Max_codeword_size = 12;
my $codeword_size = $DEFAULT_CW_SIZE;
my @dict;
my $output = "";
	
sub dprintf(@) {
#	printf @_;
}

# Format of header:
# 0-3: 32-bit filesize
# 4-5: short: lowest 9 bits (first codeword) must be 0x100 (init dictionary).
sub is_valid_lzw_file {
	my $file = shift;
	my ($size, $code);
	return -1 if -s $file < 6;                 # File must be at least 6 bytes.
	read $file, $size, 6;
	($size, $code) = unpack("VS", $size);
	return -1 if $size > 0xffffff;             # U6 files cannot exceed 16M.
	return -1 unless ($code & 0x1ff) == 0x100; # Verify first 9-bit codeword is 0x100.
	seek $file, 4, SEEK_SET;
	return $size;
}


{
	my $bits_read = 0;
	sub get_next_codeword {
		my $buf = shift;
		my $bits_needed = $codeword_size;
		my $cw = 0;
		my $offset = $bits_read % 8;
		my $byte_num = int ($bits_read / 8);
		my $bit_shift = 0;
		dprintf "$byte_num, $offset\n";
		while ($bits_needed > 0)
		{
			dprintf "length: %d byte_num: %d\n", length $$buf, $byte_num;
			warn("malformed LZW data: no end marker\n"), return undef if ($byte_num >= (length $$buf));
			my $byte = unpack("C", substr($$buf, $byte_num++, 1));
			$cw |= $byte << $bit_shift;
			$cw >>= $offset;
			dprintf "byte: %02x bit_shift: $bit_shift shifted-byte: %08x cw: %08x\n", $byte, $bit_shift, $byte << $bit_shift, $cw;
			$bits_needed -= 8 - $offset;
			$bit_shift += 8 - $offset;
			$offset = 0;
			dprintf "bits needed: $bits_needed\n";
		}
		$bits_read += $codeword_size;
		return $cw & ((1 << $codeword_size) - 1);
	}
}

sub output_root { 
	while (@_) {
		$output .= chr pop;
	}
}

sub get_string {
	my $cw = shift;
	my $cur_cw = $cw;
	my @stack;

	while ($cur_cw > 0xff) {
		push @stack, $dict[$cur_cw]{root};
		$cur_cw = $dict[$cur_cw]{codeword};
	}
	push @stack, $cur_cw;
	return @stack;
}

# args:    $buf = reference to LZW compressed file buffer (minus filesize)
# returns: $buf = reference to decompressed file as string
sub lzw_decompress_buffer {
	my $buf = shift;   # ref
	my $cw;
	my $pw;

	$dict[0x101] = 0;

	for (;;) {
		$cw = get_next_codeword($buf);
		return undef if not defined $cw;
		last if $cw == 0x101;  # Termination codeword
		if ($cw == 0x100) {
			$codeword_size = $DEFAULT_CW_SIZE;
			$cw = get_next_codeword($buf);   # Next codeword must be an ASCII character.
			output_root($cw);
			dprintf "[resetting dictionary]\n";
			undef @dict;
			$dict[0x101] = 0;
		} else {
			dprintf "cw: %08x\n", $cw;
			if ($cw < @dict) {
				# Codeword is in dictionary.
				my @stack = get_string($cw);
				my $C = $stack[-1];
				output_root(@stack);
				# Add pW+C to the dictionary
				push @dict, { root => $C, codeword => $pw };
				# Increase codeword size (if possible) if the dictionary has grown too big.
				$codeword_size++ and dprintf "[inc cw size] " if @dict >= (1 << $codeword_size) 
				                and $codeword_size < $Max_codeword_size;
			} else {
				# I don't fully understand how we can encounter a codeword not in the dictionary.
				my @stack = get_string($pw);
				my $C = $stack[-1];
				output_root(@stack);
				output_root($C);

				warn("cw != next free codeword\ncw_size: $codeword_size"), return undef if $cw != @dict;
				push @dict, { root => $C, codeword => $pw };
				$codeword_size++ and dprintf "[inc cw size] " if @dict >= (1 << $codeword_size) 
				                and $codeword_size < $Max_codeword_size;
			}
		}
		$pw = $cw;
	}
	dprintf "[end of file]\n";
	return \$output;
}

# args:    $file = LZW-compressed file
# returns: $buf  = scalar (string) representation of decompressed file, or undef on failure
sub lzw_decompress_file {
	my $file = shift;
	my $buf;

	open FH, $file or warn("unable to open $file for LZW decompression: $!\n"), return undef;
	my $size = is_valid_lzw_file(\*FH);
	warn("invalid LZW file $file!\n"), return 0 if $size < 0;
	dprintf "uncompressed size: $size\n";
	read FH, $buf, -s FH;  # plus 4 ignored bytes
	$buf = lzw_decompress_buffer(\$buf);
	return undef unless defined $buf;
	close FH;
	return $$buf;
}

1;
