package U6::MapEdit;

use strict;
use Data::Dumper;

use U6::Map qw(chunk_to_world world_to_chunk);
use U6::BookEdit;
use U6::TileEdit;
use U6::StackEdit;
use U6::Util;

# Initial map width and height; map can be resized.
my $default_map_width  = 800;
my $default_map_height = 768;

# Constructor arguments (optional): map width, map height.
sub create {
	my $obclass = shift;
	my $class = ref($obclass) || $obclass;
	my $self = {
		win => new Gtk::Dialog(),
		bookedit  => create U6::BookEdit,
		tileedit  => create U6::TileEdit,
		stackedit => create U6::StackEdit,
		animate_toggle => undef,
		objects_toggle => undef,
		grid_toggle  => undef,
		update_timer => undef,
		spin         => {},
		drawing_area => undef,
		wx => undef,
		wy => undef,
		wz => 0,
		xoff => 0,
		yoff => 0,
		buffer => undef,
	};
	bless $self, $class;

	my ($map_width, $map_height) = @_;
	$map_width  ||= $default_map_width;
	$map_height ||= $default_map_height;

	my $win = $self->{win};
	$win->set_policy( 1, 1, 0);
	$win->set_title( "Map Editor" );
	# Set default x size, because otherwise the map buttons may expand the map window.
	$win->set_default_size($map_width,-1);
	# catching destroy doesn't work this way
	$win->signal_connect( 'destroy', sub { $self->quit_application(@_); } );

	my $menubar = $self->create_menu();
	$self->{drawing_area} = new Gtk::DrawingArea();
	$self->{drawing_area}->size( $map_width, $map_height );
	my $buffer = "\0" x (($map_width + TILE_WIDTH) * ($map_height + TILE_HEIGHT));
	$self->{buffer} = \$buffer;
	$win->vbox->pack_start( $menubar, 0, 0, 0 );
	$win->vbox->pack_start( $self->{drawing_area}, 1, 1, 0 );

	my $spin = $self->{spin};
	# FIXME Put this in a loop
	$spin->{scx}{adj}  = new Gtk::Adjustment( 2, 0, 7, 1, 10, 0 );  # x superchunk adjuster
	$spin->{scx}{spin} = new Gtk::SpinButton( $spin->{scx}{adj}, 1.0, 0);
	$spin->{scy}{adj}  = new Gtk::Adjustment( 2, 0, 7, 1, 10, 0 );  # y superchunk adjuster
	$spin->{scy}{spin} = new Gtk::SpinButton( $spin->{scy}{adj}, 1.0, 0);
	$spin->{cx}{adj}   = new Gtk::Adjustment( 3, 0, 15, 1, 10, 0 ); # x chunk adjuster
	$spin->{cx}{spin}  = new Gtk::SpinButton( $spin->{cx}{adj}, 1.0, 0);
	$spin->{cy}{adj}   = new Gtk::Adjustment(10, 0, 15, 1, 10, 0 ); # y chunk adjuster
	$spin->{cy}{spin}  = new Gtk::SpinButton( $spin->{cy}{adj}, 1.0, 0);
	$spin->{tx}{adj}   = new Gtk::Adjustment( 3, 0, 7, 1, 10, 0 );  # x tile adjuster
	$spin->{tx}{spin}  = new Gtk::SpinButton( $spin->{tx}{adj}, 1.0, 0);
	$spin->{ty}{adj}   = new Gtk::Adjustment( 4, 0, 7, 1, 10, 0 );  # y tile adjuster
	$spin->{ty}{spin}  = new Gtk::SpinButton( $spin->{ty}{adj}, 1.0, 0);

	for ("scx", "scy", "cx", "cy", "tx", "ty") {
		$spin->{$_}{adj}->signal_connect('value_changed', sub {
				# We can receive a value_changed event after coordinates
				# are updated, since that also updates the spin values.
				return if $self->{spin_set_by_us}; 
				# The user changed the spin values; update our coordinates.
				# Spin values are always in range--we don't have to wrap.
				($self->{wx}, $self->{wy}) = chunk_to_world($self->spins);
				$self->{win}->set_title( sprintf("Map Editor (%03x, %03x)", $self->center_coords()) );
				::draw_map_viewport();
			} );
		$win->action_area->pack_start(new Gtk::Label($_), 0, 0, 0);
		$win->action_area->pack_start($spin->{$_}{spin}, 0, 0, 0);
	}

	# It's not necessary to manually compute the center offsets 
	# if the map is always centered when resized (as it is now).
#	$self->compute_center_offsets($map_width, $map_height);
#	$self->coords(0x11b, 0x154);
	$self->center_coords(0x134, 0x16c);

	$self->{drawing_area}->signal_connect( 'expose_event', \&::draw_map_viewport );
	$self->{drawing_area}->signal_connect( 'configure_event', sub { $self->configure(@_) } );
	# $window_map->signal_connect( 'key_release_event', \&map_keypress );

	# Check arrow keypresses in window, instead of map.  This eliminates the
	# need to focus, but prevents us from using arrows in the spin buttons.
	$win->signal_connect( 'key_press_event', sub { $self->map_keypress(@_) } );
	$self->{drawing_area}->signal_connect( 'button_press_event', \&::map_buttonpress, $spin );
	$self->{drawing_area}->signal_connect( 'button_release_event', \&::map_buttonrelease, $spin );
	#$self->{drawing_area}->signal_connect( 'key_press_event', \&map_keypress, \%spin );
	$self->{drawing_area}->set_events(['button_press_mask', 'button_release_mask', 'key_press_mask']);  # unnecessary

	$self->{drawing_area}->can_focus(1);  # can't click to focus, but can tab-to-focus

	return $self;
}

sub configure {
	my $self = shift;
	my @coords = $self->center_coords;
	my $buffer = $self->{buffer};  # ref

	my $w = $self->{drawing_area}->allocation->[2];
	my $h = $self->{drawing_area}->allocation->[3];
	$self->compute_center_offsets($w,$h);
	# Always re-center map to old center coordinates when resized.
	$self->center_coords(@coords);

	# Extend buffer.  Remember, we can overspray by up to one full tile.
	my $alloc_diff = (($w + TILE_WIDTH) * ($h + TILE_HEIGHT)) - length($$buffer);
	# Extend by requested amount, + 50k so we're not extending every time.
	substr($$buffer, -1, 0) = ("\0" x ($alloc_diff + 50000)) if $alloc_diff > 0;
	::draw_map_viewport();
}

sub compute_center_offsets {
	my $self = shift;
	my ($w, $h) = @_;
	$self->{xoff}  = int($w / TILE_WIDTH / 2);
	$self->{yoff}  = int($h / TILE_HEIGHT / 2);
}

sub create_menu {
	my $self = shift;

	my @menu = (
		{ path => '/_File',                type => '<Branch>' },
		{ path => '/File/_Save',           accelerator => '<control>S', 
		                                   callback => \&save },
		{ path => '/File/_Quit',           accelerator => '<control>Q', 
		                                   callback => \&quit_application },
		{ path => '/_View',                type => '<Branch>' },
		{ path => '/View/_Animated tiles', type => '<CheckItem>',
		                                   callback => sub { $self->toggle_animation() } },
		{ path => '/View/_Objects',        type => '<CheckItem>' },
		{ path => '/View/_Grid',           type => '<CheckItem>' },
		{ path => '/_Options',             type => '<Branch>' },
		{ path => '/Options/Edit _Terrain',type => '<CheckItem>' },
		{ path => '/_Window',              type => '<Branch>' },
		{ path => '/Window/_Book Editor',  callback => sub { $self->{bookedit} = $self->{bookedit}->show(); }},
		{ path => '/Window/_Tile Viewer',  callback => sub { $self->{tileedit}->show(); }},
		{ path => '/Window/Separator 1',   type => '<Separator>' },
		{ path => '/Window/_Go to...',     accelerator => '<control>G',
		                                   callback => sub { $self->go_to() }},
#		{ path => '/Window/_Stack Editor',  callback => sub { $self->{stackedit}->show(); }}
	);

	my $item_factory = get_main_menu($self->{win}, @menu);
	$self->{animate_toggle} = $item_factory->get_widget("/View/Animated tiles");
	$self->{objects_toggle} = $item_factory->get_widget("/View/Objects");
	$self->{grid_toggle}    = $item_factory->get_widget("/View/Grid");
	$self->{terrain_toggle} = $item_factory->get_widget("/Options/Edit Terrain");
	$self->{objects_toggle}->activate();
	
	return ( $item_factory->get_widget( '<main>' ) );
}

# Stolen from perl-gtk example program.
# args: $window widget, @menu_items in ItemFactory format
# outputs: ItemFactory [NOT menubar!]
#
# This is generic code, and should be placed in a more generic file.
sub get_main_menu {
    my ( $window, @menu_items ) = @_;

    my $accel_group  = new Gtk::AccelGroup();
    my $item_factory = new Gtk::ItemFactory( 'Gtk::MenuBar', '<main>', $accel_group );

    $item_factory->create_items( @menu_items );

    # Attach the new accelerator group to the window.
    $window->add_accel_group( $accel_group );

    # Return the item factory.
    return $item_factory;
}

sub quit_application {
	Gtk->exit( 0 );
}

sub save {
	U6::Obj::write_changes();
	U6::Map::write();
}

sub toggle_animation {
	my $self = shift;
	my $w = $self->{animate_toggle};
	my $timer = $self->{update_timer};

	if ($w->active) {
		if (not defined $timer) {
 			$self->{update_timer} = Gtk->timeout_add( 50, \&::update_game_timer, () );
		}
	} else {
		Gtk->timeout_remove( $timer ) if defined $timer;
		undef $self->{update_timer};
	}		
}

sub animation_is_active {
	my $self = shift;
	return $self->{animate_toggle}->active() && defined $self->{update_timer};
} 

sub objects_are_active {
	return $_[0]->{objects_toggle}->active();
}

sub grid_is_active {
	return $_[0]->{grid_toggle}->active();
}

sub show {
	$_[0]->{win}->show_all();
}

sub spin {
	return $_[0]->{spin};
}
sub drawing_area {
	return $_[0]->{drawing_area};
}
# Get or set the current world coordinates.
# If set, the viewport will be updated.
sub coords {
	my $self = shift;
	if (@_) {
		# Set coordinate values
		$self->{wx} = shift() & 1023;  # Wrap coords mod 1024
		$self->{wy} = shift() & 1023;
		# Set the spin values
		$self->spins(world_to_chunk($self->{wx}, $self->{wy}));
		# I don't redraw the viewport here primarily because I
		# call this function in the constructor.
	}
	# This can probably be consolidated somehow with the duplicate code in the
	# value_changed signal handler.
	$self->{win}->set_title( sprintf("Map Editor (%03x, %03x)", $self->center_coords()) );
	return ($self->{wx}, $self->{wy});
}

# Get or set the 6 coordinate spin values.
sub spins {
	# This is pretty silly now.
	my $self  = shift;
    my @spins = map { $self->{spin}->{$_}{spin} } ( "scx", "scy", "cx", "cy", "tx", "ty" );

	# Allow callers to set the spin values with passed args.
	# Notify the value_changed event that it's to ignore our updates
	# to the spin values, so we don't recurse.
	if (@_) {
		$self->{spin_set_by_us} = 1;
		for my $index (0 .. $#_) {
			$spins[$index]->set_value($_[$index]);
		}
		$self->{spin_set_by_us} = 0;
	}

    map { $_->get_value_as_int } @spins;
}

sub screen_to_world {
	my $self = shift;
	my ($x, $y) = @_;
	# World coordinates of upper-left tile
	my ($wx, $wy) = $self->coords;
	# World coordinates of specified position on screen
	# Tile limit: 1024x1024
	return (($wx + int($x / 16)) & 1023, ($wy + int($y / 16)) & 1023);
}

sub center_coords {
	my $self = shift;
	my ($x, $y, $z) = @_;
	my $xoff = $self->{xoff};
	my $yoff = $self->{yoff};
	if (defined $x and defined $y) {
		$self->coords($x - $xoff, $y - $yoff);
	}
	return ($self->{wx} + $xoff, $self->{wy} + $yoff, $self->{wz});
}

# Action performed when a key is pressed in the Map Editor window.
sub map_keypress {
	my $self = shift;
	my ($w, $e) = @_;
	use Gtk::Keysyms;

	my ($wx, $wy) = $self->coords;
	# Motion modifier: hold control to move by superchunks;
	# hold shift to move by tiles. (Shift doesn't currently work
	# on the numeric keypad.)
	my $m = 8;
	   $m *= 16 if $e->{state} & GDK_CONTROL_MASK;
	   $m /= 8  if $e->{state} & GDK_SHIFT_MASK;

	for ($e->{'keyval'}) {
		# Move by tiles on arrow keys
		$_ == $Gtk::Keysyms{'Up'}    and $wy--, next;
		$_ == $Gtk::Keysyms{'Down'}  and $wy++, next;
		$_ == $Gtk::Keysyms{'Left'}  and $wx--, next;
		$_ == $Gtk::Keysyms{'Right'} and $wx++, next;

		# Move by chunks on keys above arrow keys ;)
		$_ == $Gtk::Keysyms{'Home'}      and $wy-=$m, next;
		$_ == $Gtk::Keysyms{'End'}       and $wy+=$m, next;
		$_ == $Gtk::Keysyms{'Delete'}    and $wx-=$m, next;
		$_ == $Gtk::Keysyms{'Page_Down'} and $wx+=$m, next;

		# Move by chunks on numeric keypad
		$_ == $Gtk::Keysyms{'KP_Up'}        and $wy-=$m, next;
		$_ == $Gtk::Keysyms{'KP_Down'}      and $wy+=$m, next;
		$_ == $Gtk::Keysyms{'KP_Left'}      and $wx-=$m, next;
		$_ == $Gtk::Keysyms{'KP_Right'}     and $wx+=$m, next;
		$_ == $Gtk::Keysyms{'KP_Home'}      and $wx-=$m, $wy-=$m, next;
		$_ == $Gtk::Keysyms{'KP_End'}       and $wx-=$m, $wy+=$m, next;
		$_ == $Gtk::Keysyms{'KP_Page_Up'}   and $wx+=$m, $wy-=$m, next;
		$_ == $Gtk::Keysyms{'KP_Page_Down'} and $wx+=$m, $wy+=$m, next;

		last;  # skip the continue (redraw) block
	} continue {
		# When we return 1, it's supposed to signal the event has been handled.
		# This doesn't seem to work (in gtk1.2).  So we forcibly stop the event.
		$w->signal_emit_stop('key_press_event');
		$self->coords($wx, $wy);  # Update coordinates
		::draw_map_viewport() unless $self->animation_is_active();  # this is horribly kludgey now
		return 1;
	}
	return 0;
}

sub go_to {
	my $self = shift;
	my $dialog = new Gtk::Dialog;
	$dialog->set_title ( "Go to..." );
	my $go = new Gtk::Button 'Go';
	my $cancel = new Gtk::Button 'Cancel';
	my $ret;
	my (%adj, %spin, %hbox);

	my ($x, $y, $z) = $self->center_coords();
	$adj{x}  = new Gtk::Adjustment( $x, 0, 1023, 1, 64, 0 );
	$adj{y}  = new Gtk::Adjustment( $y, 0, 1023, 1, 64, 0 );
	$adj{z}  = new Gtk::Adjustment( $z, 0,    4, 1,  1, 0 );
	for ("x", "y", "z") {
		$spin{$_} = new Gtk::HexEntry( $adj{$_} );
		$spin{$_}->set_usize(30, 0);
		$hbox{$_} = new Gtk::HBox( 0, 0 );
		$hbox{$_}->pack_start(new Gtk::Label($_), 0, 0, 10);
		$hbox{$_}->pack_start($spin{$_}, 1, 1, 5);
		$dialog->vbox->pack_start ( $hbox{$_}, 1, 1, 5 );
	}

	$dialog->action_area->pack_start ( $go, 1, 1, 5 );
	$dialog->action_area->pack_start ( $cancel, 1, 1, 5 );

	$go->signal_connect ( clicked => sub { $ret = 1; Gtk->main_quit; 1 } );
	$cancel->signal_connect ( clicked => sub { $ret = 0; Gtk->main_quit; 1 } );
	$dialog->signal_connect ( key_press_event => sub {
				my ($w, $e) = @_;
				for ($e->{keyval}) {
					$_ == $Gtk::Keysyms{'Return'} and do {
						$ret = 1; Gtk->main_quit; return 1;
					};
					$_ == $Gtk::Keysyms{'Escape'} and do {
						$ret = 0; Gtk->main_quit; return 1;
					};
				}
				return 0;
		}
	);

	$dialog->show_all();
	# 
	$dialog->set_transient_for ( $self->{win} ) if $self->{win};
	$spin{x}->grab_focus();
#	$dialog->set_modal ( 1 );
	

	Gtk->main;
	$dialog->destroy;
	return 0 unless $ret;  # Cancel
	# Go...
	$self->center_coords($spin{x}->value, $spin{y}->value);
	::draw_map_viewport();
	return $ret;
}

1;
